/*
 * ifcon_drv.c - IFCON Communication Driver
 *
 * Copyright 2017, 2018 Sony Corporation
 * Copyright 2019, 2020 Sony Home Entertainment & Sound Products Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/freezer.h>
#include <linux/dma-mapping.h>
#include <linux/spi/spi.h>
#include <mach/mt_spi.h>
#include <mach/mt_gpio.h>
#include <cust_gpio_usage.h>
#include "ifcon_drv.h"

#define DEBUG_LOG			(0)
#define DEBUG_RINGEDGE			(0)
#define DEBUG_PACKET			(0)
#define DEBUG_SYSFS			(1)
#define EMULATE_READ_BLOCKING		(1)
#define EMULATE_TRNSMIT_ERROR		(0)

#define LOG_ERR(fmt, args...) dev_err(&ifcon_drv_dev->spi->dev, \
				"[%s]:[%d]" fmt, __func__, __LINE__, ##args)
#define LOG_WAR(fmt, args...) dev_warn(&ifcon_drv_dev->spi->dev, \
				"[%s]:[%d]" fmt, __func__, __LINE__, ##args)
#define LOG_INF(fmt, args...) dev_info(&ifcon_drv_dev->spi->dev, fmt, ##args)
#if DEBUG_LOG
#define LOG_DBG(fmt, args...) dev_info(&ifcon_drv_dev->spi->dev, fmt, ##args)
#else
#define LOG_DBG(fmt, args...) dev_dbg(&ifcon_drv_dev->spi->dev, fmt, ##args)
#endif

#define IFCON_DRV_MINORS		(32) /* total number of minor number */

#define IFCON_DRV_RING_BUFF_SIZE	(2048) /* size of ring buffer */

/* Self status */
#define IFCON_DRV_CTRL_START_CONDITION	(0x50)
#define IFCON_DRV_CTRL_NORMAL		(0x55)
#define IFCON_DRV_CTRL_RESENDING	(0x5E)
#define IFCON_DRV_CTRL_RESEND_REQ	(0x5F)
#define IFCON_DRV_CTRL_BROKEN_PACKET	(0xFF)
#define IFCON_DRV_CTRL_INVALID		(0x00)
/* Another side status */
#define IFCON_DRV_RX_START_CONDITION	(0x30)
#define IFCON_DRV_RX_NORMAL		(0x33)
#define IFCON_DRV_RX_RESENDING		(0x3E)
#define IFCON_DRV_RX_RESEND_REQUEST	(0x3F)
/* Flow ctrl */
#define IFCON_DRV_FLOW_FOR_CHECK	(0x01)
#define IFCON_DRV_FLOW_NOT_FULL		(0x00)
#define IFCON_DRV_FLOW_NEAR_FULL	(0x80)
#define IFCON_DRV_FLOW_FULL		(0xC0)
#define IFCON_DRV_FLOW_PACKET_LOST	(0xE0)
#define IFCON_DRV_FLOW_ALL		(0xF0)
/* Counter */
#define IFCON_DRV_COUNTER_MAX		(255)
#define IFCON_DRV_WARN_COUNT_MAX	(20)

/* Module status */
#define IFCON_DRV_STAT_INIT		(0)
#define IFCON_DRV_STAT_CONNECT		(1)

/* Suspend status */
#define IFCON_DRV_POWER_NORMAL		(0)
#define IFCON_DRV_POWER_SUSPEND		(1)

/* If communication reset */
#if EMULATE_READ_BLOCKING
#define IFCON_DRV_READ_NO_BLOCK		(0)
#define IFCON_DRV_READ_BLOCKING		(1)
#endif

/* display frame data */
#define IFCON_DRV_SEL_RX		(0)
#define IFCON_DRV_SEL_TX		(1)
#define IFCON_DRV_SEL_TX_RX		(2)

/* main process cycle */
#if USE_256FRAME
#define IFCON_DRV_NORMAL_CYCLE		(2 * HZ / 100)	/* 20ms cycle */
#define IFCON_DRV_READCK_CYCLE		(2000 / 2)	/* 20ms/2 us */
#else
#define IFCON_DRV_NORMAL_CYCLE		(5 * HZ / 100)	/* 50ms cycle */
#define IFCON_DRV_READCK_CYCLE		(5000 / 2)	/* 50ms/2 us */
#endif
#define IFCON_DRV_LOSS_OUT_COUNT	(100)	/* 20ms * 100 cycle */

#define IFCON_DRV_RESERVED_SIZE		(5)
#define IFCON_DRV_NEAR_FULL_SIZE	(2 * IFCON_DRV_RW_DATA_SIZE)

static DECLARE_BITMAP(minors, IFCON_DRV_MINORS);

struct ifcon_drv_data_t {
	dev_t			devt;
	dev_t			major;
	struct spi_device	*spi;
	struct cdev		cdev;
	struct class		*class;
	struct device		*dev;
	struct task_struct	*kthread_tsk;
	struct mutex		state_lock;
	wait_queue_head_t	rxwait;
	unsigned int		users;
	u8			count_stat;
	u8			count_flow;
};

static struct ifcon_drv_data_t *ifcon_drv_dev;

/* Tx/Rx SPI com frame */
struct ifcon_drv_frame {
	u8	status;		/* Communication status */
	u8	flow_ctrl;	/* Communication flow control */
	u8	count;		/* Communication Counter[0-255] */
	u8	reserved[IFCON_DRV_RESERVED_SIZE];
	u8	fixed_data[IFCON_DRV_FIXED_DATA_SIZE];	/* Fixed data area */
	u8	data[IFCON_DRV_RW_DATA_SIZE];	/* Access for User space */
	u16	check_sum;	/* Check sum */
};
#define IFCON_SPI_FRAME_SIZE	(sizeof(struct ifcon_drv_frame))

static struct ifcon_drv_frame *tx_frame;	/* Access for SPI register */
static dma_addr_t tx_phyadr;
static struct ifcon_drv_frame *rx_frame;	/* Access for SPI register */
static dma_addr_t rx_phyadr;

/* Tx/Rx ring buffer */
struct ifcon_drv_ring_buffer {
	u8	buffer[IFCON_DRV_RING_BUFF_SIZE];
	int	rptr;		/* Read pointer for ring buffer */
	int	wptr;		/* Write pointer for ring buffer */
};

static struct ifcon_drv_ring_buffer tx_ring;
static struct ifcon_drv_ring_buffer rx_ring;
static	int	tx_xptr;	/* Tx - transmit pointer, for resend used */
static	int	tx_vptr;	/* Tx - verify pointer, for resend request */

static int ifcon_drv_status = IFCON_DRV_STAT_INIT;
static int ifcon_drv_power = IFCON_DRV_POWER_NORMAL;
#if EMULATE_READ_BLOCKING
static int ifcon_drv_rblock = IFCON_DRV_READ_NO_BLOCK;
#endif
static int ifcon_drv_boot = 0;	/* booted flag */
static int ifcon_drv_loss = 0;	/* Not received count */

#if DEBUG_SYSFS
static int ifcon_status = 0;
module_param_named(ifcon_status, ifcon_status, int, S_IWUSR | S_IRUGO );
#endif

/*--- Dump Tx/Rx buffer --------------------------------------------*/

/* for error case */
static void display_frame_data(u8 mode)
{
	int i, p;
	char buf[64];

	if ((mode == IFCON_DRV_SEL_TX_RX) || (mode == IFCON_DRV_SEL_TX)) {
		LOG_INF("IFCON_DRV_SEL_TX:\n");
		p = 0;
		for (i = 0; i < IFCON_DRV_FRAME_ALL_SIZE; i++) {
			if ((i % 16) == 0)
				p += snprintf(&buf[p], 5, "%03x:", i);
			p += snprintf(&buf[p], 4, " %02x",
						*((char *)tx_frame + i));
			if ((i % 16) == 15) {
				pr_info("%s\n", buf);
				p = 0;
			}
		}
		if (p) pr_info("%s\n", buf);

	}
	if ((mode == IFCON_DRV_SEL_TX_RX) || (mode == IFCON_DRV_SEL_RX)) {
		LOG_INF("IFCON_DRV_SEL_RX:\n");
		p = 0;
		for (i = 0; i < IFCON_DRV_FRAME_ALL_SIZE; i++) {
			if ((i % 16) == 0)
				p += snprintf(&buf[p], 5, "%03x:", i);
			p += snprintf(&buf[p], 4, " %02x",
						*((char *)rx_frame + i));
			if ((i % 16) == 15) {
				pr_info("%s\n", buf);
				p = 0;
			}
		}
		if (p) pr_info("%s\n", buf);
	}
}

static void warn_rx_status(s8 *str)
{
	if (ifcon_drv_dev->count_stat < IFCON_DRV_WARN_COUNT_MAX) {
		ifcon_drv_dev->count_stat++;
		LOG_WAR("%s 0x%02x size:0x%02x cnt:0x%02x\n", str,
			rx_frame->status, rx_frame->data[0], rx_frame->count);
	}
}

static void warn_rx_tx_flow(s8 *str)
{
	if (ifcon_drv_dev->count_flow < IFCON_DRV_WARN_COUNT_MAX) {
		ifcon_drv_dev->count_flow++;
		LOG_WAR("%s rx:0x%02x tx:0x%02x cnt:0x%02x\n", str,
		    rx_frame->flow_ctrl, tx_frame->flow_ctrl, rx_frame->count);
	}
}

#if DEBUG_PACKET
#define STR_SIZE (3)
#define RESEND (0x0E)
/* for debug */
static void debug_display_packet_size(struct ifcon_drv_frame *frame, u8 mode)
{
	int sz = 0;
	u8 str[] = "Tx";

	if (mode == IFCON_DRV_SEL_RX)
		snprintf(str, STR_SIZE, "Rx");

	if ((((frame->status & RESEND) == RESEND) || (frame->flow_ctrl != 0))
	   && (frame->data[sz] == 0))
		LOG_INF("%s status:0x%02x flow:0x%02x cnt:%d\n",
		  str, frame->status, frame->flow_ctrl, frame->count);
	while ((frame->data[sz] > 0)
	    && (frame->data[sz] < IFCON_DRV_RW_DATA_SIZE)) {
		if (sz == 0)
			LOG_INF("%s status:0x%02x flow:0x%02x cnt:%d\n", str,
				frame->status, frame->flow_ctrl, frame->count);
		LOG_INF(" size:0x%02x\n", frame->data[sz]);
		sz += frame->data[sz];
	}
}
#endif

#if DEBUG_SYSFS
static void display_internal_state(void)
{
	int i, p;
	char buf[64];

	pr_info("ifcon_drv_status = %02x\n", ifcon_drv_status);
	pr_info("ifcon_drv_power  = %02x\n", ifcon_drv_power);
	pr_info("ifcon_drv_rblock = %02x\n", ifcon_drv_rblock);
	pr_info("rx_frame.status  = %02x\n", rx_frame->status);
	pr_info("rx_frame.flow_.. = %02x\n", rx_frame->flow_ctrl);
	pr_info("rx_wptr:%03x rx_rptr:%03x\n", rx_ring.wptr, rx_ring.rptr);
	pr_info("tx_frame.status  = %02x\n", tx_frame->status);
	pr_info("tx_frame.flow_.. = %02x\n", tx_frame->flow_ctrl);
	pr_info("tx_wptr:%03x tx_rptr:%03x\n", tx_ring.wptr, tx_ring.rptr);
	pr_info("tx_xptr:%03x tx_vptr:%03x\n", tx_xptr, tx_vptr);

	LOG_INF("RX_RING:\n");
	p = 0;
	for (i = 0; i < IFCON_DRV_RING_BUFF_SIZE; i++) {
		if ((i % 16) == 0)
			p += snprintf(&buf[p], 5, "%03x:", i);
		p += snprintf(&buf[p], 4, " %02x", rx_ring.buffer[i]);
		if ((i % 16) == 15) {
			pr_info("%s\n", buf);
			p = 0;
		}
	}
	if (p) pr_info("%s\n", buf);

	LOG_INF("TX_RING:\n");
	p = 0;
	for (i = 0; i < IFCON_DRV_RING_BUFF_SIZE; i++) {
		if ((i % 16) == 0)
			p += snprintf(&buf[p], 5, "%03x:", i);
		p += snprintf(&buf[p], 4, " %02x", tx_ring.buffer[i]);
		if ((i % 16) == 15) {
			pr_info("%s\n", buf);
			p = 0;
		}
	}
	if (p) pr_info("%s\n", buf);
}
#endif

/******************************************************************************/
/*!	Request Resend
 *	The request of Resend is requested from another side.
 *	When word status is IFCON_DRV_RX_RESEND_REQUEST.
 */
/******************************************************************************/
static void resend_req(void)
{
	tx_ring.rptr = tx_vptr;
	tx_xptr = tx_ring.rptr;
}

/******************************************************************************/
/*!	Verifying OK
 *	This is verify the sending data is OK
 *	When word 1 is not 0x5F(resend request)
 */
/******************************************************************************/
static void verify_ok(void)
{
	tx_vptr = tx_xptr;
	tx_xptr = tx_ring.rptr;
}

/* Get flow control code */
static u8 get_flow_code(struct ifcon_drv_ring_buffer *ring)
{
	int rest = 0;

	/* rest writable buffer size */
	if (ring->rptr <= ring->wptr)
		rest = IFCON_DRV_RING_BUFF_SIZE - (ring->wptr - ring->rptr);
	else
		rest = ring->rptr - ring->wptr;
	/* Flow control code */
	if (rest < IFCON_DRV_RW_DATA_SIZE)
		return IFCON_DRV_FLOW_FULL;
	else if (rest < IFCON_DRV_NEAR_FULL_SIZE)
		return IFCON_DRV_FLOW_NEAR_FULL;
	else
		return IFCON_DRV_FLOW_NOT_FULL;
}

/*--- Check ctrl area / Check sum --------------------------------------------*/
static u16 calc_check_sum(u8 *buf)
{
	u8 i = 0;
	u16 sum = 0;

	for (i = 0; i < (IFCON_DRV_FRAME_ALL_SIZE - 2); i++)
		sum = buf[i] + sum;
	/* complement on two */
	sum = ~sum + 1;
	/* check sum value */
	return sum;
}

static inline int compare_check_sum(u8 *buf, u16 checksum)
{
	u16 sum = calc_check_sum(buf);
	/* checksum value check */
	if (checksum == sum) {
		return IFCON_DRV_OK;
	} else {
		LOG_ERR("check sum error! 0x%x -> calc: 0x%x size: %d\n",
			checksum, sum, rx_frame->data[0]);
		return IFCON_DRV_NG;
	}
}

static inline void ifcon_hw_reset(void)
{
	LOG_ERR("Ifcon force reset!\n");
	mt_set_gpio_out(GPIO_IFCON_RESET_PIN, GPIO_OUT_ZERO);
	usleep_range(10000, 11000);
	mt_set_gpio_out(GPIO_IFCON_RESET_PIN, GPIO_OUT_ONE);
}

static int check_ctrl_data(void)
{
	int ret = IFCON_DRV_NG;
	u32 total_size = 0;

	while ((total_size < IFCON_DRV_RW_DATA_SIZE)
	    && (rx_frame->data[total_size] > 0))
		total_size += rx_frame->data[total_size];
	/* Check data size */
	if (total_size > IFCON_DRV_RW_DATA_SIZE) {
		warn_rx_status("check_ctrl_data: size > RW_DATA_SIZE error!:");
	} else {
		ret = IFCON_DRV_OK;
		/* Controll Area: status */
		if (rx_frame->status == IFCON_DRV_RX_START_CONDITION) {
			ifcon_drv_dev->count_stat = 0;
			tx_frame->status = IFCON_DRV_CTRL_NORMAL;
			ifcon_drv_status = IFCON_DRV_STAT_CONNECT;
			warn_rx_status("check_ctrl_data: Start up:");
		} else if (rx_frame->status == IFCON_DRV_RX_NORMAL) {
			tx_frame->status = IFCON_DRV_CTRL_NORMAL;
		} else if (rx_frame->status == IFCON_DRV_RX_RESENDING) {
			tx_frame->status = IFCON_DRV_CTRL_NORMAL;
			warn_rx_status("check_ctrl_data: Resending:");
		} else if (rx_frame->status == IFCON_DRV_RX_RESEND_REQUEST) {
			tx_frame->status = IFCON_DRV_CTRL_RESENDING;
			warn_rx_status("check_ctrl_data: Resend req:");
		} else if (rx_frame->status == IFCON_DRV_CTRL_BROKEN_PACKET) {
			tx_frame->status = IFCON_DRV_CTRL_RESEND_REQ;
			ret = IFCON_DRV_NG;	/* discard data. */
			warn_rx_status("check_ctrl_data: Broken packet!:");
		} else {
			ret = IFCON_DRV_NG;	/* discard data. */
			warn_rx_status("check_ctrl_data: Status Invalid!:");
		}
		/* Another side status: request resend ? */
		if (ret == IFCON_DRV_OK) {
			if (rx_frame->status == IFCON_DRV_RX_RESEND_REQUEST)
				resend_req();
			else
				verify_ok();
		}
		/* Self status: request resend ? */
		if (tx_frame->status == IFCON_DRV_CTRL_RESEND_REQ)
			ret = IFCON_DRV_NG;	/* discard data. */
		/* invalid data ? */
		if ((rx_frame->flow_ctrl & IFCON_DRV_FLOW_FOR_CHECK) != 0) {
			ret = IFCON_DRV_NG;	/* discard data. */
			warn_rx_status("IFCON_DRV_FLOW_FOR_CHECK!: 0x01 stat:");
		}
		/* If communication reset */
		if ((rx_frame->status == IFCON_DRV_CTRL_INVALID) &&
		    (rx_frame->count == 0) &&
		    (rx_frame->data[0] == 0) &&
		    (rx_frame->check_sum == 0)) {
			if (ifcon_drv_boot == 0
			 && ifcon_drv_loss == IFCON_DRV_LOSS_OUT_COUNT) {
				ifcon_hw_reset();
				ifcon_drv_boot = 1;
			}
			ifcon_drv_loss++;
			warn_rx_status("Ifcon spi reset!:");
			tx_frame->status = IFCON_DRV_CTRL_START_CONDITION;
			tx_frame->flow_ctrl = IFCON_DRV_FLOW_NOT_FULL;
			ifcon_drv_status = IFCON_DRV_STAT_INIT;
#if EMULATE_READ_BLOCKING
			ifcon_drv_rblock = IFCON_DRV_READ_NO_BLOCK;
#endif
			wake_up_interruptible(&ifcon_drv_dev->rxwait);
			ret = IFCON_DRV_NG;
		} else {
			ifcon_drv_boot = 1;
		}
		/* check rx sum */
		if ((ret == IFCON_DRV_OK)
		 && (ifcon_drv_status > IFCON_DRV_STAT_INIT)) {
			ret = compare_check_sum((u8 *)rx_frame,
						rx_frame->check_sum);
			if ((ret != IFCON_DRV_OK) && (rx_frame->data[0] != 0)) {
				tx_frame->status = IFCON_DRV_CTRL_RESEND_REQ;
				display_frame_data(IFCON_DRV_SEL_RX);
			}
		}
	}

	return ret;
}

/*--- Access for Rx Ring buffer -------------------------------------------*/

/* Read from ring buffer */
static int rx_ring_read(char __user *buf, int count)
{
	int wptr_temp, rptr_temp;
	int sz_total, sz;
	int rest_cap;

	wptr_temp = rx_ring.wptr;
	rptr_temp = rx_ring.rptr;
	sz_total = rx_ring.buffer[rptr_temp];	/*1st byte=size*/
	sz = sz_total;
	rest_cap = IFCON_DRV_RING_BUFF_SIZE - rptr_temp;

	if (rptr_temp == wptr_temp)
		return 0;
	if ((sz <= 0) || (sz > IFCON_DRV_RW_DATA_SIZE)) {
		LOG_ERR("pointer size error! sz:%d buf:0x%x\n",
					sz, (unsigned int)buf);
		return IFCON_DRV_NG;
	}
	if (count < sz) {
		LOG_WAR("RX_BUFF_SIZE_ERROR sz:%d count:%d\n", sz, count);
		return -IFCON_DRV_RX_BUFF_SIZE_ERROR;
	}
	if (rptr_temp >= IFCON_DRV_RING_BUFF_SIZE) { /* Pointer inside check */
		LOG_ERR("rx_ring.rptr error! %d\n", wptr_temp);
		return IFCON_DRV_NG;
	}

	/* Read from ring buffer Rx */
	if (sz >= rest_cap) {
		if (rptr_temp < wptr_temp) {
			LOG_ERR("err! r:%d w:%d\n", rptr_temp, wptr_temp);
			return IFCON_DRV_NG;
		}
#if DEBUG_RINGEDGE
		LOG_INF("(sz >= rest_cap)sz:%d rest_cap:%d\n", sz, rest_cap);
#endif
		if (copy_to_user(buf, &rx_ring.buffer[rptr_temp], rest_cap)) {
			LOG_ERR("RX copy_to_user\n");
			return -EFAULT;
		}
		rptr_temp = 0;
		sz -= rest_cap;
		buf = (char *)(buf + rest_cap);
	}
	if (copy_to_user(buf, &rx_ring.buffer[rptr_temp], sz)) {
		LOG_ERR("RX copy_to_user\n");
		return -EFAULT;
	}
	/* Read Pointer increment */
	barrier();
	rx_ring.rptr = rptr_temp + sz;

	return sz_total;
}

/* Write to ring buffer */
static int rx_ring_write(char *buf, int count)
{
	int rptr_temp, wptr_temp;
	int rest_cap, rest;
	int sz;

	rptr_temp = rx_ring.rptr;
	wptr_temp = rx_ring.wptr;
	rest_cap = IFCON_DRV_RING_BUFF_SIZE - wptr_temp;
	sz = count;
	rest = 0;

	if (count != *buf) {
		LOG_ERR("(count != *buf) error! c:0x%x b:0x%x\n", sz, *buf);
		return IFCON_DRV_NG;
	}
	if ((sz <= 0) || (sz > IFCON_DRV_RW_DATA_SIZE)) {
		LOG_ERR("pointer size error! count:%d buf:0x%x\n",
					sz, (unsigned int)buf);
		return IFCON_DRV_NG;
	}
	if (wptr_temp >= IFCON_DRV_RING_BUFF_SIZE) { /* Pointer inside check */
		LOG_ERR("rx_ring.wptr error! %d\n", wptr_temp);
		return IFCON_DRV_NG;
	}

	/* rest writable buffer size */
	if (rptr_temp <= wptr_temp)
		rest = IFCON_DRV_RING_BUFF_SIZE - (wptr_temp - rptr_temp);
	else
		rest = rptr_temp - wptr_temp;
	/* Flow control */
	if (rest < IFCON_DRV_RW_DATA_SIZE) {
		LOG_ERR("BUFF_FULL! size:%d rest:%d\n", sz, rest);
		return -IFCON_DRV_BUFF_FULL;
	}
	/* Write to ring buffer Tx */
	if (sz >= rest_cap) {
#if DEBUG_RINGEDGE
		LOG_INF("(sz >= rest_cap)sz:%d rest_cap:%d\n", sz, rest_cap);
#endif
		/* Write to end edge */
		memcpy(&rx_ring.buffer[wptr_temp], buf, rest_cap);
		wptr_temp = 0;
		sz -= rest_cap;
		buf = (char *)(buf + rest_cap);
	}
	memcpy(&rx_ring.buffer[wptr_temp], buf, sz);
	/* Write Pointer increment */
	barrier();
	rx_ring.wptr = wptr_temp + sz;

	return count;
}

/* Write to ring buffer Rx */
static int copy_from_frame_to_rx_ring(char *buf)
{
	int sz_total, sz;
	char *buf_next;

	sz_total = 0;
	sz = *buf;	/*1st byte=size*/
	buf_next = buf;

	/* Read from ring buffer Tx */
	while ((sz > 0) && (sz_total < (IFCON_DRV_RW_DATA_SIZE - 1))) {
		if (*buf_next == 0x00)
			break;
		sz = rx_ring_write(buf_next, *buf_next);
		if (sz > 0) {
			sz_total += sz;
			buf_next = buf + sz_total;
			if (((sz_total + (*buf_next)) > IFCON_DRV_RW_DATA_SIZE)
			  && (sz_total < IFCON_DRV_RW_DATA_SIZE)) {
				LOG_ERR("error! total:%d\n",
					sz_total + (*buf_next));
				break;
			}
		}
	}

	if (sz < 0) {	/* error case */
		sz_total = sz;
		if (sz != -IFCON_DRV_BUFF_FULL)
			display_frame_data(IFCON_DRV_SEL_TX_RX);
	}

	if (rx_ring.rptr != rx_ring.wptr)
		wake_up_interruptible(&ifcon_drv_dev->rxwait);

	return sz_total;
}

static int get_rx_frame(void)
{
	int status = 0;

	/* Process for loading Rx data */
	status = check_ctrl_data(); /* check rx data */
	/* Write to ring buffer Rx */
	if (status == IFCON_DRV_OK)
		status = copy_from_frame_to_rx_ring(rx_frame->data);

	return status;
}

/*--- Access for Tx Ring buffer -------------------------------------------*/

/* Read from ring buffer */
static int tx_ring_read(char *buf, int count, int used_sz)
{
	int wptr_temp, rptr_temp;
	int sz_total, sz;
	int rest_cap;

	wptr_temp = tx_ring.wptr;
	rptr_temp = tx_ring.rptr;
	sz_total = tx_ring.buffer[rptr_temp];	/*1st byte=size*/
	sz = sz_total;
	rest_cap = IFCON_DRV_RING_BUFF_SIZE - rptr_temp;

	if (rptr_temp == wptr_temp)
		return 0;
	if ((sz <= 0) || (sz > IFCON_DRV_RW_DATA_SIZE)) {
		LOG_ERR("pointer size error! sz:%d buf:0x%x\n",
					sz, (unsigned int)buf);
		return IFCON_DRV_NG;
	}
	if (sz + used_sz > IFCON_DRV_RW_DATA_SIZE)
		return 0;	/* Not enter this frame */
	if (count < sz) {
		LOG_WAR("TX_BUFF_SIZE_ERROR sz:%d count:%d\n", sz, count);
		return -IFCON_DRV_RX_BUFF_SIZE_ERROR;
	}
	if (rptr_temp >= IFCON_DRV_RING_BUFF_SIZE) { /* Pointer inside check */
		LOG_ERR("tx_ring.rptr error! %d\n", rptr_temp);
		return IFCON_DRV_NG;
	}

	/* Read from ring buffer Rx */
	if (sz >= rest_cap) {
		if (rptr_temp < wptr_temp) {
			LOG_ERR("err! r:%d w:%d\n", rptr_temp, wptr_temp);
			return IFCON_DRV_NG;
		}
#if DEBUG_RINGEDGE
		LOG_INF("(sz >= rest_cap)sz:%d rest_cap:%d\n", sz, rest_cap);
#endif
		memcpy(buf, &tx_ring.buffer[rptr_temp], rest_cap);
		rptr_temp = 0;
		sz -= rest_cap;
		buf = (char *)(buf + rest_cap);
	}
	memcpy(buf, &tx_ring.buffer[rptr_temp], sz);
	/* Read Pointer increment */
	barrier();
	tx_ring.rptr = rptr_temp + sz;

	return sz_total;
}

/* Write to ring buffer */
static int tx_ring_write(const char __user *buf, int count)
{
	int wptr_temp, vptr_temp;
	int rest_cap, rest;
	int sz;

	wptr_temp = tx_ring.wptr;
	vptr_temp = tx_vptr;
	rest_cap = IFCON_DRV_RING_BUFF_SIZE - wptr_temp;
	sz = count;
	rest = 0;

	if (count != *buf) {
		LOG_ERR("(count != *buf) error! c:0x%x b:0x%x\n", sz, *buf);
		return IFCON_DRV_NG;
	}
	if ((sz <= 0) || (sz > IFCON_DRV_RW_DATA_SIZE)) {
		LOG_ERR("pointer size error! count:%d buf:0x%x\n",
					sz, (unsigned int)buf);
		return IFCON_DRV_NG;
	}
	if (wptr_temp >= IFCON_DRV_RING_BUFF_SIZE) { /* Pointer inside check */
		LOG_ERR("tx_ring.wptr error! %d\n", wptr_temp);
		return IFCON_DRV_NG;
	}

	/* rest writable buffer size */
	if (vptr_temp <= wptr_temp)
		rest = IFCON_DRV_RING_BUFF_SIZE - (wptr_temp - vptr_temp);
	else
		rest = vptr_temp - wptr_temp;
	/* Flow control */
	if (rest < IFCON_DRV_RW_DATA_SIZE) {
		LOG_ERR("BUFF_FULL! size:%d rest:%d\n", sz, rest);
		return -IFCON_DRV_BUFF_FULL;
	}

	/* Write to ring buffer Tx */
	if (sz >= rest_cap) {
#if DEBUG_RINGEDGE
		LOG_INF("(sz >= rest_cap)sz:%d rest_cap:%d\n", sz, rest_cap);
#endif
		/* Write to end edge */
		memcpy(&tx_ring.buffer[wptr_temp], buf, rest_cap);
		wptr_temp = 0;
		sz -= rest_cap;
		buf = (char *)(buf + rest_cap);
	}
	memcpy(&tx_ring.buffer[wptr_temp], buf, sz);
	/* Write Pointer increment */
	barrier();
	tx_ring.wptr = wptr_temp + sz;

	return count;
}

/* Read from ring buffer Tx */
static int copy_from_tx_ring_to_frame(char *buf)
{
	int sz_total, sz;
	char *buf_next;

	sz = IFCON_DRV_RW_DATA_SIZE;
	sz_total = 0;
	buf_next = buf;

#if DEBUG_RINGEDGE
	LOG_INF("tx_wptr:%d tx_rptr:%d\n", tx_ring.wptr, tx_ring.rptr);
	LOG_INF("rx_wptr:%d rx_rptr:%d\n", rx_ring.wptr, rx_ring.rptr);
	LOG_INF("tx_xptr:%d tx_vptr:%d\n", tx_xptr, tx_vptr);
#endif
	/* Read from ring buffer Tx */
	while (sz > 0) {
		sz = tx_ring_read(buf_next,
			(IFCON_DRV_RW_DATA_SIZE - sz_total), sz_total);
		if (sz > 0)
			sz_total += sz;
		buf_next = buf + sz_total;
	}

	return sz_total;
}

static int set_tx_frame(void)
{
	int status = 0;
	u8 flow_check = 0;	/* flow check */
	u8 rx_flow;

	rx_flow = rx_frame->flow_ctrl & IFCON_DRV_FLOW_ALL; /*Another side */

	/* Set controll Area: status ->	check_ctrl_data) */
	/* Set controll Area: count  ->	set_counter() */
	/* Set controll Area: flow */
	if (rx_flow == IFCON_DRV_FLOW_FULL) {	/* (0xC0) */
		flow_check = IFCON_DRV_FLOW_FOR_CHECK;	/* (0x01) */
		warn_rx_tx_flow("detect FLOW_FULL!:");
	} else if (rx_flow == IFCON_DRV_FLOW_NEAR_FULL) {	/* (0x80) */
		if ((tx_frame->flow_ctrl & IFCON_DRV_FLOW_FOR_CHECK) == 0)
			flow_check = IFCON_DRV_FLOW_FOR_CHECK; /* alternately */
		warn_rx_tx_flow("detect FLOW_NEAR_FULL!:");
	} else if (rx_flow == IFCON_DRV_FLOW_PACKET_LOST) {	/* (0xE0) */
		resend_req();
		tx_frame->status = IFCON_DRV_CTRL_RESENDING;
		warn_rx_tx_flow("detect FLOW_PACKET_LOST!:");
	} else { /* if (rx_flow == IFCON_DRV_FLOW_NOT_FULL) { */ /* (0x00) */
		if (tx_frame->flow_ctrl == 0)
			ifcon_drv_dev->count_flow = 0;
	}
	tx_frame->flow_ctrl = get_flow_code(&rx_ring) | flow_check;
	if ((tx_frame->flow_ctrl & IFCON_DRV_FLOW_ALL) != 0)
		warn_rx_tx_flow("Tx FLOW_CTRL:");
	/* Set data from ring buffer Tx */
	memset(tx_frame->data, 0, IFCON_DRV_RW_DATA_SIZE); /* clear packet */
	if ((flow_check != IFCON_DRV_FLOW_FOR_CHECK)
	&& (ifcon_drv_status > IFCON_DRV_STAT_INIT))
		status = copy_from_tx_ring_to_frame(tx_frame->data);
	if (status < 0)
		LOG_ERR("set_tx_frame() Fatal error! status:%d\n", status);
	/* Set check sum tx */
	tx_frame->check_sum = calc_check_sum((u8 *)tx_frame);

	return status;
}

/*--- Access for SPI device --------------------------------------------------*/
#if EMULATE_TRNSMIT_ERROR
extern void get_random_bytes(void *buf, int nbytes);

static int random(void)
{
	int rd;

	get_random_bytes(&rd, sizeof(int));
	return rd;
}
#endif

/* The method of transferring SPI data that using spi_***() APIs
   are based on kernel/drivers/misc/mediatek/spi/mt8590/spi-dev.c
*/
static inline int
sync_write_and_read(struct ifcon_drv_data_t *spidev, int len)
{

	struct spi_transfer trn = {
	/* from Documentation/spi/spidev
	To make a full duplex request, provide both rx_buf and tx_buf for the
	same transfer.	It's even OK if those are the same buffer. */
		.tx_buf	= tx_frame,
		.rx_buf	= rx_frame,
		.len	= len,
		.tx_dma	= tx_phyadr,
		.rx_dma	= rx_phyadr,
	};
	struct spi_message msg;
	int status;

#if EMULATE_TRNSMIT_ERROR
	u8 *pos = NULL;
	u8 bup;
	int eps;

	if ((random() & 0xFF) == 0) {
		LOG_DBG("emulation : send data broken !!!\n");
		eps = (random() & 0x7fff) % IFCON_SPI_FRAME_SIZE;
		pos = (u8 *)tx_frame;
		bup = pos[eps];
		pos[eps] = (random() & 0xff);
	}
#endif

	spi_message_init(&msg);
	spi_message_add_tail(&trn, &msg);

	if (spidev->spi == NULL)
		status = -ESHUTDOWN;
	else {
		msg.is_dma_mapped = 1;

		dma_sync_single_for_device(NULL,
			tx_phyadr, IFCON_SPI_FRAME_SIZE, DMA_BIDIRECTIONAL);

		status = spi_sync(spidev->spi, &msg);

		dma_sync_single_for_cpu(NULL,
			rx_phyadr, IFCON_SPI_FRAME_SIZE, DMA_BIDIRECTIONAL);

		if (status == 0)
			status = msg.actual_length;
	}

#if EMULATE_TRNSMIT_ERROR
	if (pos != NULL)
		pos[eps] = bup;

	if ((random() & 0xFF) == 0) {
		LOG_DBG("emulation : receive data broken !!!\n");
		eps = (random() & 0x7fff) % IFCON_SPI_FRAME_SIZE;
		pos = (__u8 *)rx_frame;
		pos[eps] = (random() & 0xff);
	} else if ((random() & 0x3FFF) == 0) {
		LOG_DBG("emulation : no receive data !!!\n");
		memset((__u8 *)rx_frame, 0, IFCON_SPI_FRAME_SIZE);
	}
#endif

	return status;
}

static void set_counter(void)
{
	if (tx_frame->count >= IFCON_DRV_COUNTER_MAX)
		tx_frame->count = 0;
	else
		tx_frame->count++;
}

static void spi_communication(struct ifcon_drv_data_t *ifcon_drv_dev)
{
	int status = 0;

	status = set_tx_frame();
#if DEBUG_PACKET
	debug_display_packet_size(tx_frame, IFCON_DRV_SEL_TX);
#endif
	status = sync_write_and_read(ifcon_drv_dev, IFCON_DRV_FRAME_ALL_SIZE);
	if (status != IFCON_DRV_FRAME_ALL_SIZE)
		LOG_ERR("sync_write_and_read() error! status:%d\n", status);
#if DEBUG_PACKET
	debug_display_packet_size(rx_frame, IFCON_DRV_SEL_RX);
#endif
	status = get_rx_frame();

	set_counter();	/* Process after Tx */
}

/*--- 50ms cycle main routine (10ms for update) ---------------------------*/
static int spi_comm_thread(void *arg)
{
	while (!kthread_should_stop()) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(IFCON_DRV_NORMAL_CYCLE);

#if DEBUG_SYSFS
		if (ifcon_status) {
			ifcon_status = 0;
			display_internal_state();
		}
#endif

		if (ifcon_drv_power == IFCON_DRV_POWER_SUSPEND)
			continue; /* for suspend */

		mutex_lock(&ifcon_drv_dev->state_lock);
		spi_communication((struct ifcon_drv_data_t *)arg);
		mutex_unlock(&ifcon_drv_dev->state_lock);
	}
	return 0;
}

static void spi_comm_reset(void)
{
	mutex_lock(&ifcon_drv_dev->state_lock);

	/* Initialize ring buffer */
	tx_xptr = 0;
	tx_vptr = 0;
	memset(&tx_ring, 0, sizeof(tx_ring));
	memset(&rx_ring, 0, sizeof(rx_ring));

	/* init communication */
	tx_frame->status = IFCON_DRV_CTRL_START_CONDITION;
	tx_frame->flow_ctrl = IFCON_DRV_FLOW_NOT_FULL;
	ifcon_drv_status = IFCON_DRV_STAT_INIT;

	mutex_unlock(&ifcon_drv_dev->state_lock);
}

/*--- System call ---------------------------------------------------------*/
static ssize_t ifcon_drv_read(struct file *filp, char __user *buf,
					size_t count, loff_t *offset)
{
	ssize_t status = 0;

	LOG_DBG("ifcon_drv_read()\n");
#if EMULATE_READ_BLOCKING
	ifcon_drv_rblock = IFCON_DRV_READ_BLOCKING;
	while ((rx_ring.rptr == rx_ring.wptr
	     || ifcon_drv_status < IFCON_DRV_STAT_CONNECT)
	    && ifcon_drv_rblock == IFCON_DRV_READ_BLOCKING
	    && ifcon_drv_power != IFCON_DRV_POWER_SUSPEND) {
		usleep_range(IFCON_DRV_READCK_CYCLE,
			     IFCON_DRV_READCK_CYCLE * 11 / 10);
	}
	ifcon_drv_rblock = IFCON_DRV_READ_NO_BLOCK;
#endif
	if (ifcon_drv_power == IFCON_DRV_POWER_SUSPEND)
		return 0;

	/* If communication reset */
	if ((ifcon_drv_status < IFCON_DRV_STAT_CONNECT) &&
	    (ifcon_drv_dev->count_stat < IFCON_DRV_WARN_COUNT_MAX)) {
		ifcon_drv_dev->count_stat++;
		LOG_ERR(" com reset :IFCON_DRV_NOT_CONNECT error!\n");
		return -IFCON_DRV_NOT_CONNECT;
	}

	/* Read from ring buffer Rx */
	status = (ssize_t)rx_ring_read(buf, (int)count);

	LOG_DBG("ifcon_drv_read() status:%d\n", status);
	return status;
}

static ssize_t ifcon_drv_write(struct file *filp, const char __user *buf,
					size_t count, loff_t *offset)
{
	ssize_t status = 0;
	u8 tmp[IFCON_DRV_RW_DATA_SIZE];

	LOG_DBG("ifcon_drv_write() count 0x%x\n", count);
	if (count > IFCON_DRV_RW_DATA_SIZE) {
		LOG_ERR(" :count > IFCON_DRV_RW_DATA_SIZE error!\n");
		return -EMSGSIZE;
	}

	if (copy_from_user(tmp, (void *)buf, count)) {
		LOG_ERR("ifcon_drv_write(): error copy_from_user\n");
		return -EFAULT;
	}

	status = (ssize_t)tx_ring_write(tmp, (int)count);

	LOG_DBG("ifcon_drv_write() status:%d\n", status);
	return status;
}


static long ifcon_drv_ioctl(struct file *filp, unsigned int cmd,
						unsigned long arg)
{
	int err = 0;
	u32 missing = 0;
	struct st_ifcon_fixed_data_param_def st_param;

	LOG_DBG("ifcon_drv_ioctl()\n");
	/* Check type and command number */
	if (_IOC_TYPE(cmd) != IFCON_DRV_IOC_MAGIC)
		return -ENOTTY;

	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg,
							_IOC_SIZE(cmd));
	if (err == 0 && _IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void __user *)arg,
							_IOC_SIZE(cmd));
	if (err)
		return -EFAULT;

	missing = copy_from_user(&st_param, (void *)arg, sizeof(st_param));
	if (missing != 0)
		return -EFAULT;
	/* Check size */
	if ((st_param.offset + st_param.sz) > IFCON_DRV_FIXED_DATA_SIZE)
		return -EFAULT;

	switch (cmd) {
	case IFCON_DRV_IOC_WRITE_FIXED:
		missing = copy_from_user(&tx_frame->fixed_data[st_param.offset],
									st_param.data, st_param.sz);
		break;
	case IFCON_DRV_IOC_READ_FIXED:
		missing = copy_to_user(st_param.data,
									&rx_frame->fixed_data[st_param.offset], st_param.sz);
		break;
	default:
		break;
	}

	LOG_DBG("ifcon_drv_ioctl() missing:%d\n", (unsigned int)missing);
	if (missing != 0)
		return -EFAULT;
	return 0;
}

static long ifcon_drv_compat_ioctl(struct file *filp, unsigned int cmd,
							unsigned long arg)
{
	return ifcon_drv_ioctl(filp, cmd, arg);
}

static unsigned int ifcon_drv_poll(struct file *filp, poll_table *ptbl)
{
	struct ifcon_drv_data_t *ddt;
	unsigned int mask = 0;

	ddt = filp->private_data;

	LOG_DBG("ifcon_drv_poll()\n");
	if (ddt == NULL)
		return -EBADFD;

	if (rx_ring.rptr == rx_ring.wptr)
		poll_wait(filp, &ddt->rxwait, ptbl);

	if (rx_ring.rptr != rx_ring.wptr)
		mask |= POLLIN | POLLRDNORM;

	LOG_DBG("ifcon_drv_poll() status:0x%x\n", mask);
	return mask;
}

static int ifcon_drv_open(struct inode *inode, struct file *filp)
{
	LOG_INF("ifcon_drv_open()\n");

	/* Only 1 user */
	if (ifcon_drv_dev->users > 0) {
		LOG_ERR("Already open! %d\n", ifcon_drv_dev->users);
		return IFCON_DRV_NG;
	}

	ifcon_drv_dev->users++;
	filp->private_data = ifcon_drv_dev;
	nonseekable_open(inode, filp);

	/* reset spi tx/rx buffer */
	spi_comm_reset();

	LOG_DBG("ifcon_drv_open() exit\n");
	return 0;
}

static int ifcon_drv_release(struct inode *inode, struct file *filp)
{
	LOG_INF("ifcon_drv_release()\n");
	filp->private_data = NULL;
	ifcon_drv_dev->users--;

	LOG_DBG("ifcon_drv_release() exit\n");
	return 0;
}

/*-------------------------------------------------------------------------*/
static const struct file_operations ifcon_drv_fops = {
	.owner =	THIS_MODULE,
	.read =		ifcon_drv_read,
	.write =	ifcon_drv_write,
	.unlocked_ioctl = ifcon_drv_ioctl,
	.compat_ioctl = ifcon_drv_compat_ioctl,
	.poll =		ifcon_drv_poll,
	.open =		ifcon_drv_open,
	.release =	ifcon_drv_release,
};

/*-------------------------------------------------------------------------*/
static int ifcon_drv_suspend(struct spi_device *spi, pm_message_t msg)
{
	LOG_INF("ifcon_drv_suspend()\n");
	ifcon_drv_power = IFCON_DRV_POWER_SUSPEND;
	kthread_stop(ifcon_drv_dev->kthread_tsk);

	/* to IFCON stop mode */
	mt_set_gpio_out(GPIO_IFCON_BD_REQ_PIN, GPIO_OUT_ZERO);

	LOG_DBG("ifcon_drv_suspend() exit\n");
	return 0;
}

static int ifcon_drv_resume(struct spi_device *spi)
{
	LOG_INF("ifcon_drv_resume()\n");
	/* to IFCON wake mode */
	mt_set_gpio_out(GPIO_IFCON_BD_REQ_PIN, GPIO_OUT_ONE);

	ifcon_drv_dev->kthread_tsk = kthread_run(spi_comm_thread,
					(void *)ifcon_drv_dev, "spi_com");
	if (IS_ERR(ifcon_drv_dev->kthread_tsk)) {
		LOG_ERR("error! ifcon_drv_dev->kthread_tsk: 0x%x\n",
				(unsigned int)ifcon_drv_dev->kthread_tsk);
		return IFCON_DRV_NG;
	}
	ifcon_drv_power = IFCON_DRV_POWER_NORMAL;
	LOG_DBG("ifcon_drv_resume() exit\n");
	return 0;
}

static int ifcon_drv_probe(struct spi_device *spi)
{
	int status = 0;
	u32 minor = 0;

	LOG_INF("ifcon_drv_probe()\n");

	/* Allocate dma buffer */
	tx_frame = (struct ifcon_drv_frame *)dma_alloc_coherent(
						NULL, IFCON_SPI_FRAME_SIZE,
						&tx_phyadr, GFP_KERNEL);
	if (!tx_frame) {
		pr_err("tx_frame dma_alloc error! %p\n", tx_frame);
		status = -ENOMEM;
		goto out_err_exit;
	}
	rx_frame = (struct ifcon_drv_frame *)dma_alloc_coherent(
						NULL, IFCON_SPI_FRAME_SIZE,
						&rx_phyadr, GFP_KERNEL);
	if (!rx_frame) {
		pr_err("rx_frame dma_alloc error! %p\n", rx_frame);
		status = -ENOMEM;
		goto out_free_txbuf;
	}
	/* Initialize the driver data */
	ifcon_drv_dev->spi = spi;
	mutex_init(&ifcon_drv_dev->state_lock);
	init_waitqueue_head(&ifcon_drv_dev->rxwait);

	/* If we can allocate a minor number, hook up this device.
	 * Reusing minors is fine so long as udev or mdev is working.
	 */
	minor = find_first_zero_bit(minors, IFCON_DRV_MINORS);
	if (minor < IFCON_DRV_MINORS) {
		ifcon_drv_dev->devt = MKDEV(MAJOR(ifcon_drv_dev->major), minor);
		set_bit(minor, minors);

		spi_set_drvdata(spi, ifcon_drv_dev);
	} else {
		LOG_ERR("no minor number available!\n");
		status = -ENODEV;
		goto out_free_rxbuf;
	}

	/* to IFCON wake mode */
	mt_set_gpio_mode(GPIO_IFCON_BD_REQ_PIN, GPIO_IFCON_BD_REQ_PIN_M_GPIO);
	mt_set_gpio_out(GPIO_IFCON_BD_REQ_PIN, GPIO_OUT_ONE);
	mt_set_gpio_dir(GPIO_IFCON_BD_REQ_PIN, GPIO_DIR_OUT);

	/* reset spi tx/rx buffer */
	spi_comm_reset();

	/* Create & wakeup own thread */
	ifcon_drv_dev->kthread_tsk = kthread_run(spi_comm_thread,
					(void *)ifcon_drv_dev, "spi_com");
	if (IS_ERR(ifcon_drv_dev->kthread_tsk)) {
		LOG_ERR("error! ifcon_drv_dev->kthread_tsk: 0x%x\n",
				(unsigned int)ifcon_drv_dev->kthread_tsk);
		return IFCON_DRV_NG;
	}

	LOG_INF("ifcon_drv_probe() cs:%d status:%d\n",
				spi->chip_select, status);
	return status;

out_free_rxbuf:
	dma_free_coherent(NULL, IFCON_SPI_FRAME_SIZE, rx_frame, rx_phyadr);
	rx_frame  = NULL;
	rx_phyadr = 0;
out_free_txbuf:
	dma_free_coherent(NULL, IFCON_SPI_FRAME_SIZE, tx_frame, tx_phyadr);
	tx_frame  = NULL;
	tx_phyadr = 0;
out_err_exit:
	return status;
}

static int ifcon_drv_remove(struct spi_device *spi)
{
	LOG_INF("ifcon_drv_remove()\n");
	kthread_stop(ifcon_drv_dev->kthread_tsk);

	spi_set_drvdata(spi, NULL);
	ifcon_drv_dev->spi = NULL;

	device_destroy(ifcon_drv_dev->class, ifcon_drv_dev->devt);
	clear_bit(MINOR(ifcon_drv_dev->devt), minors);

	dma_free_coherent(NULL, IFCON_SPI_FRAME_SIZE, rx_frame, rx_phyadr);
	rx_frame  = NULL;
	rx_phyadr = 0;
	dma_free_coherent(NULL, IFCON_SPI_FRAME_SIZE, tx_frame, tx_phyadr);
	tx_frame  = NULL;
	tx_phyadr = 0;

	/* to IFCON stop mode */
	mt_set_gpio_out(GPIO_IFCON_BD_REQ_PIN, GPIO_OUT_ZERO);

	LOG_DBG("ifcon_drv_remove() exit\n");
	return 0;
}

/*-------------------------------------------------------------------------*/
struct spi_device_id ifcon_drv_id_table = {"ifcon_drv_spi", 0};

static struct spi_driver ifcon_driver_st = {
	.driver = {
		.name =		IFCON_DRV_DEVNAME,
		.bus =		&spi_bus_type,
		.owner =	THIS_MODULE,
	},
	.probe =	ifcon_drv_probe,
	.remove =	ifcon_drv_remove,
	.suspend =	ifcon_drv_suspend,
	.resume =	ifcon_drv_resume,
	.id_table =	&ifcon_drv_id_table,
};

/*-------------------------------------------------------------------------*/
static int __init ifcon_drv_init(void)
{
	int status = 0;

	pr_info("ifcon_drv_init()\n"); /* dev_info() is not available. */
	/* Allocate driver data */
	ifcon_drv_dev = kzalloc(sizeof(struct ifcon_drv_data_t), GFP_KERNEL);
	if (!ifcon_drv_dev) {
		pr_err("dev kzalloc error! %p\n", ifcon_drv_dev);
		return -ENOMEM;
	}
	/* Register SPI device */
	status = alloc_chrdev_region(&ifcon_drv_dev->major, 0, 1, "spi");
	if (status) {
		pr_err("alloc_chrdev_region() error:0x%x\n", status);
		goto out_free;
	}
	cdev_init(&ifcon_drv_dev->cdev, &ifcon_drv_fops);
	ifcon_drv_dev->cdev.owner = THIS_MODULE;

	status = cdev_add(&ifcon_drv_dev->cdev, ifcon_drv_dev->major, 1);
	if (status) {
		pr_err("cdev_add() error:0x%x\n", status);
		goto out_unregister;
	}

	/* Create ifcon_drv device class */
	ifcon_drv_dev->class = class_create(THIS_MODULE, IFCON_DRV_DEVNAME);
	if (IS_ERR(ifcon_drv_dev->class)) {
		status = PTR_ERR(ifcon_drv_dev->class);
		pr_err("class_create() error:0x%x\n", status);
		goto out_cdev_del;
	}

	ifcon_drv_dev->dev = device_create(ifcon_drv_dev->class, NULL,
				ifcon_drv_dev->major, NULL, IFCON_DRV_DEVNAME);
	if (IS_ERR(ifcon_drv_dev->dev)) {
		status = PTR_ERR(ifcon_drv_dev->dev);
		pr_err("device_create() error:0x%x\n", status);
		goto out_class_destroy;
	}

	/* Register SPI driver */
	/*spi_register_board_info(spi0_devs, ARRAY_SIZE(spi0_devs));*/
	/* -> kernel\arch\arm\mach-mt8590\BBBAR1_linux\core\model.c */

	status = spi_register_driver(&ifcon_driver_st);
	if (status < 0) {
		pr_err("spi_register_driver() error:0x%x\n", status);
		goto out_class_destroy;
	}

	return status;

out_class_destroy:
	class_destroy(ifcon_drv_dev->class);
out_cdev_del:
	cdev_del(&ifcon_drv_dev->cdev);
out_unregister:
	unregister_chrdev_region(ifcon_drv_dev->major, 1);
out_free:
	kfree(ifcon_drv_dev);

	return status;
}
module_init(ifcon_drv_init);

static void __exit ifcon_drv_exit(void)
{
	pr_info("ifcon_drv_exit()\n");
	spi_unregister_driver(&ifcon_driver_st);
	class_destroy(ifcon_drv_dev->class);
	cdev_del(&ifcon_drv_dev->cdev);
	unregister_chrdev_region(ifcon_drv_dev->major, 1);
	kfree(ifcon_drv_dev);
}
module_exit(ifcon_drv_exit);

MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("IfCon driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("spi:ifcon_drv");
